/* episim.c - Main control program, command line processing, etc.
   Author: Robert D. Vincent
*/
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <math.h>
#include <limits.h>
#include <string.h>

#include "stdjunk.h"
#include "graph.h"
#include "iafs.h"
#include "mersenne.h"

#define POLICY_PERIODIC 0
#define POLICY_RANDOM 1
#define POLICY_ONESHOT 2

int
main(int argc, char **argv)
{
  struct hh_net *net_p;
  graph_t S;
  fp_t *Ie;
  int tflag = 0;
  fp_t t;
  int i;
  int j;
  int k = 10;
  fp_t p = 0.01;
  time_t tm;
  fp_t Toff = 90;
  fp_t Ton = 10;
  fp_t T = 2000.0;
  fp_t Is = 16.5;
  char *opts = NULL;
  int Nsf=-1;
  int Nsl=-1;
  unsigned long net_seed = 0;
  unsigned long sig_seed = 1;

  /* Output files */
  FILE *fps;
  FILE *fpK;
  FILE *fpNa;
  FILE *fpv;
  FILE *fpIsyn;
  FILE *fpG;

  int gtype = 's';		/* graph type */
  int qflag = 0;
  int action;
  int policy_type = POLICY_PERIODIC;
  fp_t pinterval = 0.001;	/* Printing interval, seconds */
  int kprint;
  fp_t xdelay = 1e10;           /* very big number */
  int xtrigger = 0;

  char outfile[1024];
  char stem[1000] = {""};

  for (i = 1; i < argc; i++) {
    char *argp = argv[i];
    if (*argp == '-') {
      argp++;
      switch (*argp) {
      case 'X':
        argp++;
        xdelay = atof(argp);
        break;
      case 'q':
        qflag++;
        break;
      case 's':
        strncpy(stem, argp+1, sizeof(stem)-1);
        break;
      case 't':
	tflag++;
	break;
      case 'G':
	argp++;
	switch (*argp) {
	case 'r':
	  gtype = 'r';		/* rectangular lattice */
	  break;
	case 'l':
	  gtype = 'l';		/* line */
	  break;
	default:
	  gtype = 's';		/* ring */
	  break;
	}
	break;
      case 'k':
	argp++;
	k = atoi(argp);
	break;

      case 'R':
	argp++;
	net_seed = atol(argp);
	break;

      case 'r':
	argp++;
	sig_seed = atol(argp);
	break;

      case 'p':
	argp++;
	p = atof(argp);
	if (p < 0.0 || p > 1.0) {
	  fprintf(stderr, "Illegal value (%f) for p\n", p);
	  exit(-1);
	}
	break;

      case 'T':
	argp++;
	if (argp[0] == 'o' && argp[1] == 'n') {
	  argp += 2;
	  Ton = atof(argp);
	}
	else if (argp[0] == 'o' && argp[1] == 'f' && argp[2] == 'f') {
	  argp += 3;
	  Toff = atof(argp);
	}
	else {
	  T = atof(argp);
	}
	break;

      case 'I':
	argp++;
	if (argp[0] == 's') {
	  argp++;
	  Is = atof(argp);
	}
	break;

      case 'N':
	argp++;
	if (argp[0] == 's' && argp[1] == 'l') {
	  argp += 2;
	  Nsl = atoi(argp);
	}
	else if (argp[0] == 's' && argp[1] == 'f') {
	  argp += 2;
	  Nsf = atoi(argp);
	}
	break;

      case 'S':
	policy_type = POLICY_RANDOM;
	break;

      case 'O':
	policy_type = POLICY_ONESHOT;
	break;

      case 'P':
	argp++;
	pinterval = atof(argp);
	break;

      default:
	break;
      }
    }
    else {
      opts = argp;
    }
  }

  if (Nsf==-1) {
    Nsf = 20;
  }
  if (Nsl==-1) {
    Nsl = Nsf+20;
  }
  if (tflag) {
    time(&tm);
    fprintf(stderr, "1: %s", ctime(&tm));
  }

  sprintf(outfile, "s%s.txt", stem);
  fps = fopen(outfile, "w");

  if (!qflag) {
    sprintf(outfile, "v%s.txt", stem);
    fpv = fopen(outfile, "w");

    sprintf(outfile, "K%s.txt", stem);
    fpK = fopen(outfile, "w");
    
    sprintf(outfile, "N%s.txt", stem);
    fpNa = fopen(outfile, "w");

    sprintf(outfile, "Isyn%s.txt", stem);
    fpIsyn = fopen(outfile, "w");
  }
  else {
    fpv = NULL;
    fpK = NULL;
    fpNa = NULL;
    fpIsyn = NULL;
  }

  sprintf(outfile, "graph%s.txt", stem);
  fpG = fopen(outfile, "w");

  fprintf(fps, "%% k = %d\n", k);
  fprintf(fps, "%% Network seed is %ld\n", net_seed);
  fprintf(fps, "%% Signal seed is %ld\n", sig_seed);
  fprintf(fps, "%% p = %f\n", p);
  fprintf(fps, "%% Ton = %g\n", Ton);
  fprintf(fps, "%% Toff = %g\n", Toff);
  fprintf(fps, "%% T = %f\n", T);
  fprintf(fps, "%% Is = %f\n", Is);
  fprintf(fps, "%% Nsl = %d\n", Nsl);
  fprintf(fps, "%% Nsf = %d\n", Nsf);
  switch (policy_type) {
  case POLICY_RANDOM:
    fprintf(fps, "%% Random policy\n");
    break;
  case POLICY_ONESHOT:
    fprintf(fps, "%% Single pulse\n");
    break;
  default:
    fprintf(fps, "%% Periodic policy\n");
    break;
  }

  fprintf(fps, "%% command line: ");
  for (i = 0; i < argc; i++) {
    fprintf(fps, "%s ", argv[i]);
  }
  fprintf(fps, "\n");

  init_genrand(net_seed);

  net_p = init_hh(opts);

  fprintf(fps, "%% simulating %d neurons\n", net_p->n);

  init_genrand(net_seed);

  if (gtype == 'r') {
    int n, m;
    m = k;
    n = net_p->n / m;
    S = rect(n, m, p, net_p->Pmax, net_p->Psd);
    fprintf(fps, "%% Using rectangular %dx%d lattice\n", n, m);
  }
  else if (gtype == 'l') {
    S = line(net_p->n, k, p, net_p->Pmax, net_p->Psd);
    fprintf(fps, "%% Using 1D line graph\n");
  }
  else {
    S = ring(net_p->n, k, p, net_p->Pmax, net_p->Psd);
    fprintf(fps, "%% Using ring graph\n");
  }

  graph_print(fpG, S, net_p->inh);
  fclose(fpG);

  if (tflag) {
    time(&tm);
    fprintf(stderr, "2: %s", ctime(&tm));
  }

  init_genrand(sig_seed);

  Ie = (fp_t *) malloc(sizeof(fp_t) * net_p->n);

  kprint = floor((pinterval / net_p->dT) + 0.5);

  for (i = 0, t = 0.0; t <= T; i++, t += net_p->dT) {
    action = 0;
    /* Generate a Poisson train of stimulation pulses.
     */
    if (policy_type == POLICY_RANDOM) {
      if (((double) rand() / RAND_MAX) < ((double)Ton / (Toff + Ton))) {
	action = 1;
      }
    }
    /* Generate a single pulse with width Ton and delay Toff
     */
    else if (policy_type == POLICY_ONESHOT) {
      if (t >= Toff && t <= Toff+Ton) {
	action = 1;
      }
    }
    /* Generate a square wave which is off for Toff then on for Ton
     */
    else {
      fp_t x = fmod(t, Toff+Ton);
      if (x >= Toff) {
	action = 1;
      }
    }

    if (action) {
      for (j = 0; j < net_p->n; j++) {
	if (j >= Nsf && j <= Nsl) {
	  Ie[j] = Is;
	}
	else {
	  Ie[j] = 0.0;
	}
      }
    }
    else {
      for (j = 0; j < net_p->n; j++) {
	Ie[j] = 0.0;
      }
    }

    if (t > xdelay && xtrigger == 0) {
      /* Remove AMPA completely */
      //net_p->gAMPA = 0;
      /* Remove NMDA completely */
      net_p->gNMDA = 0;
      xtrigger = 0;
    }

    step_hh(net_p, S, t, Ie);

    /* Just print out the spikes */
    for (j = 0; j < net_p->n; j++) {
      if (net_p->F[j]) {
	fprintf(fps, "%f %d\n", t, j);
      }
    }

    if ((i % kprint) != 0) {
      continue;
    }

    if ((i % 10000) == 0) {
      fp_t sum = 0.0;
      fprintf(stdout, "%f ", t);
      for (j=0;j<net_p->n; j++) {
	sum += net_p->X[j];
      }
      fprintf(stdout, "%f ", sum / net_p->n);

      fprintf(stdout, "%f ", net_p->chvars[1*N_CHVARS+I_CoK]);

      sum = 0.0;
      for (j=0;j<net_p->n; j++) {
	sum += net_p->chvars[(j*N_CHVARS)+I_CoK];
      }
      fprintf(stdout, "%f ", sum / net_p->n);

      sum = 0.0;
      for (j=0;j<net_p->n; j++) {
	sum += net_p->chvars[j*N_CHVARS+I_B];
      }
      fprintf(stdout, "%f ", sum / net_p->n);

      fprintf(stdout, "\n");
    }

    if (qflag) {
      continue;
    }

    /* Print out the voltages.
     */
    fprintf(fpv, "%f ", t);
    for (j = 0; j < net_p->n; j++) {
      fprintf(fpv, "%.3f ", net_p->chvars[j*N_CHVARS+I_Vm]);
    }
    fprintf(fpv, "\n");

    /* Print the potassium concentrations.
     */
    fprintf(fpK, "%f ", t);
    for (j = 0; j < net_p->n; j++) {
      fprintf(fpK, "%.3f %.3f ", 
              net_p->chvars[j*N_CHVARS+I_CiK], 
              net_p->chvars[j*N_CHVARS+I_CoK]);
    }
    fprintf(fpK, "\n");

    /* Print the sodium concentrations.
     */
    fprintf(fpNa, "%f ", t);
    for (j = 0; j < net_p->n; j++) {
      fprintf(fpNa, "%.3f %.3f ", 
              net_p->chvars[j*N_CHVARS+I_CiNa], 
              net_p->chvars[j*N_CHVARS+I_CoNa]);
    }
    fprintf(fpNa, "\n");

    /* Print the full synaptic current field. 
     */
    fprintf(fpIsyn, "%f ", t);
    for (j = 0; j < net_p->n; j++) {
      fprintf(fpIsyn, "%.4g ", net_p->Isyn[j]);
    }
    fprintf(fpIsyn, "\n");
  }

  if (tflag) {
    time(&tm);
    fprintf(stderr, "3: %s", ctime(&tm));
  }
  fclose(fps);
  if (!qflag) {
    fclose(fpv);
    fclose(fpK);
    fclose(fpNa);
    fclose(fpIsyn);
  }
  return (0);
}
