/* ga.c - Golomb/Amitai cell 
 * Author: Robert D. Vincent
 */
#include <math.h>
#include <stdio.h>

#include "stdjunk.h"
#include "graph.h"
#include "iafs.h"
#include "ga.h"

// Steady-state values using our "typical" parameters
#define V0 -67.553

static const fp_t phi = 10.0;

/* INa (fast sodium) */

/* From Golomb et al 2006 */
static const fp_t theta_h = -45;		/* mV */
static const fp_t sigma_h = -7;		/* mV */

static const fp_t theta_m = -30.0;		/* mV */
static const fp_t sigma_m = 9.5;		/* mV */
static const fp_t theta_th = -40.5;		/* mV */
static const fp_t sigma_th = -6;		/* mV */

/* INaP (persistent sodium)  */
/* From Golomb et al. 2006: */
static const fp_t theta_p = -47;		// mV
static const fp_t sigma_p = 3;		// mV

/* IKdr (potassium delayed rectifier) */
static const fp_t theta_n = -33;
static const fp_t sigma_n = 10.0;
static const fp_t theta_tn = -27.0;
static const fp_t sigma_tn = -15.0;

// IKA (A-type potassium current)
static const fp_t theta_a = -50.0;
static const fp_t sigma_a = 20.0;
static const fp_t theta_b = -80.0;
static const fp_t sigma_b = -6.0;
static const fp_t tau_b = 15.0;

// IKSlow
static const fp_t theta_z = -39;
static const fp_t sigma_z = 5;
static const fp_t tau_z = 75;

void
init_ga(fp_t *y)
{
  y[I_Vm] = V0;
  y[I_USER+0] = sigmoid(V0, theta_n, sigma_n);
  y[I_USER+1] = sigmoid(V0, theta_h, sigma_h);
  y[I_USER+2] = sigmoid(V0, theta_z, sigma_z);
  y[I_USER+3] = sigmoid(V0, theta_b, sigma_b);
}

/* Update Golomb/Amitai neuron.
 */
void
update_ga(struct hh_net *net_p, /* Network structure */
          fp_t *y,              /* State variables */
	  int i, 		/* Unit index */
	  fp_t i_inf, 		/* Synaptic target */
	  fp_t g_sum, 		/* Synaptic conductance */
	  fp_t *ik,
	  fp_t *ina,
          fp_t *dy)
{
  // I leak
  fp_t E_L = net_p->EL;		/* -70 mV */
  fp_t E_Na = net_p->ENa;	/* 55.0 mV */
  fp_t E_K = net_p->EK;		/* -90 mV */

  fp_t g_tmp;

  fp_t tau_n, n_inf;		/* Delayed-rectifier K+ (Kdr) */

  fp_t a_inf, b_inf;		/* A-type K+ */

  fp_t tau_h, h_inf;		/* Fast Na+ */
  fp_t m_inf;

  fp_t p_inf;			/* Persistent Na+ (NaP) */

  fp_t z_inf;			/* Slow K+ (Kslow) */

  fp_t tau_v, v_inf;

  fp_t v = y[I_Vm];
  fp_t n = y[I_USER+0];
  fp_t h = y[I_USER+1];
  fp_t z = y[I_USER+2];
  fp_t b = y[I_USER+3];

  fp_t i_now;
  fp_t i_tmp;

  i_now = 0;

  *ik = 0;
  *ina = 0;

  /*** Fast Na+ current
   */
  m_inf = sigmoid(v, theta_m, sigma_m);

  g_tmp = net_p->gNax * pow(m_inf, 3) * h;
  g_sum += g_tmp;
  i_inf += g_tmp * E_Na;
  i_tmp = g_tmp * (v - E_Na);
  i_now += i_tmp;
  *ina += i_tmp;

  /*** Persistent Na+ current
   */
  p_inf = sigmoid(v, theta_p, sigma_p);
  
  g_tmp = net_p->gNaPx * p_inf;
  g_sum += g_tmp;
  i_inf += g_tmp * E_Na;
  i_tmp = g_tmp * (v - E_Na);
  i_now += i_tmp;
  *ina += i_tmp;

  /*** Delayed-rectifier K+ 
   */
  g_tmp = net_p->gKdrx * pow(n, 4);
  g_sum += g_tmp;
  i_inf += g_tmp * E_K;
  i_tmp = g_tmp * (v - E_K);
  i_now += i_tmp;
  *ik += i_tmp;

  /*** A-type K+ 
   * Used in G&A 1997 only, max conductance set to zero in 
   * G.et al. 2006.
   */
  a_inf = sigmoid(v, theta_a, sigma_a);

  g_tmp = net_p->gKAx * pow(a_inf, 3) * b;
  g_sum += g_tmp;
  i_inf += g_tmp * E_K;
  i_tmp = g_tmp * (v - E_K);
  i_now += i_tmp;
  *ik += i_tmp;

  /*** Slow K+
   */
  g_tmp = net_p->gKslowx * z;
  g_sum += g_tmp;
  i_inf += g_tmp * E_K;
  i_tmp = g_tmp * (v - E_K);
  i_now += i_tmp;
  *ik += i_tmp;

  /* Leak */
  g_sum += net_p->gLx;
  i_inf += net_p->gLx * E_L;
  i_now += net_p->gLx * (v - E_L);

  v_inf = i_inf / g_sum;
  tau_v = net_p->cm / g_sum;
  
  n_inf = sigmoid(v, theta_n, sigma_n);
  tau_n = (1.0 + 5.0 * sigmoid(v, theta_tn, sigma_tn)) / phi;

  h_inf = sigmoid(v, theta_h, sigma_h);
  tau_h = (1.0 + 7.5 * sigmoid(v, theta_th, sigma_th)) / phi;
  
  b_inf = sigmoid(v, theta_b, sigma_b);

  z_inf = sigmoid(v, theta_z, sigma_z);

  /* Integrate ! */
  dy[I_Vm] = (v_inf - v) / tau_v;
  dy[I_USER+0] = (n_inf - n) / tau_n;
  dy[I_USER+1] = (h_inf - h) / tau_h;
  dy[I_USER+2] = (z_inf - z) / tau_z;
  dy[I_USER+3] = (b_inf - b) / tau_b;

  /* Save results back to state */
  net_p->Im[i] = i_now;
}

