#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include "stdjunk.h"
#include "rnd.h"
#include "graph.h"

edge_t
graph_get_edge(graph_t g, int i, int j)
{
  edge_t e_p;

  for (e_p = g->e[j]; e_p != NULL; e_p = e_p->link) {
    if (e_p->src == i) {
      return (e_p);
    }
  }
  return (NULL);
}

void
graph_add_edge(graph_t g, int i, int j, fp_t w)
{
  edge_t e_p = (edge_t) malloc(sizeof(struct edge));
  if (e_p != NULL) {
    e_p->link = g->e[j];
    e_p->src = i;
    e_p->weight = w;
    g->e[j] = e_p;
  }
}

void
graph_clr_edge(graph_t g, int i, int j)
{
  edge_t e_p;
  edge_t p_p;

  for (p_p = NULL, e_p = g->e[j]; e_p != NULL; p_p = e_p, e_p = e_p->link) {
    if (e_p->src == i) {
      if (p_p == NULL) {
	g->e[j] = e_p->link;
      }
      else {
	p_p->link = e_p->link;
      }
      free(e_p);
      return;
    }
  }
  fprintf(stderr, "warning: edge not found\n");
}

graph_t
graph_alloc(int n)
{
  int i;

  graph_t g = (graph_t) malloc(sizeof(struct graph) + (n - 1) * sizeof(struct edge *));
  if (g == NULL) {
    return NULL;
  }

  g->n = n;

  for (i = 0; i < n; i++) {
    g->e[i] = NULL;
  }
  return (g);
}

graph_t smallworld(graph_t g, fp_t p)
{
  int ne, nr;
  int np = 0, nn = 0;
  int i, j;
  int t;
  edge_t e_p;
  struct rewire_list {
    struct rewire_list *link;
    int src;
    int dst;
    fp_t weight;
  } *rw_list_p, *rw_elem_p;

  ne = 0;
  nr = 0;


  rw_list_p = NULL;
  for (i = 0; i < g->n; i++) {
    for (j = 0; j < g->n; j++) {
      if ((e_p = graph_get_edge(g, i, j)) != NULL) {
	ne++;
	if (rnd_uni() < p) {
	  rw_elem_p = malloc(sizeof(struct rewire_list));

	  rw_elem_p->src = i;
	  rw_elem_p->dst = j;
	  rw_elem_p->weight = e_p->weight;
	  rw_elem_p->link = rw_list_p;
	  rw_list_p = rw_elem_p;
	}
      }
    }
  }

  while ((rw_elem_p = rw_list_p) != NULL) {

    rw_list_p = rw_elem_p->link;

    /* Rewire this connection 
     */
    do {
      /* Uniformly randomly choose a new target. 
       */
      t = (int) floor(rnd_uni() * g->n);
    } while (t == rw_elem_p->src || 
	     graph_get_edge(g, rw_elem_p->src, t) != NULL);

    /* Delete the existing edge.
     */
    graph_clr_edge(g, rw_elem_p->src, rw_elem_p->dst);

    /* Add the new edge.
     */
    graph_add_edge(g, rw_elem_p->src, t, rw_elem_p->weight);

    free(rw_elem_p);
    nr++;
  }
  fprintf(stderr, "Rewired %d of %d edges\n", nr, ne);

  for (i = 0; i < g->n; i++) {
    for (e_p = g->e[i]; e_p != NULL; e_p = e_p->link) {
      if (e_p->weight >= 0.0) {
	np++;
      }
      else {
	nn++;
      }
    }
  }
  fprintf(stderr, "%d Positive, %d Negative weights\n", np, nn);
  return (g);
}

/* build a small world network:
 * 1. each of n nodes is connected to k neighbors
 * 2. with probability p, an edge is rewired to a random node
 * 
 * Note that self-loops and redundant edges are disallowed.
 * 
 * Pmax and Psd describe a Gaussian distribution of weights.
 */
graph_t ring(int n, int k, fp_t p, fp_t Pmax, fp_t Psd)
{
  int i, j, t;
  graph_t g = graph_alloc(n);

  /* Don't crash in the unlikely event that n < k
   */
  if (n < k) {
    k = n - 1;
    if (k < 0 || n < 0) {
      return NULL;
    }
  }

  /* These nested loops connect up the basic lattice.
   */
  for (i = 0; i < n; i++) {
    for (j = i-(k/2); j <= i+(k/2); j++) {
      if (j < 0) 
	t = j + n;
      else if (j >= n) 
	t = j - n;
      else
	t = j;
      
      if (t != i) {
	if (Psd >= 0.0) {
	  graph_add_edge(g, i, t, (Psd * rnd_nrm()) + Pmax);
	}
	else {
	  graph_add_edge(g, i, t, Pmax);
	}
      }
    }
  }

  return smallworld(g, p);
}

/* build a rectangular lattice
 * each of n nodes is connected to 8 neighbors
 * 
 * Pmax and Psd describe a Gaussian distribution of weights.
 */
graph_t rect(int n, int m, fp_t p, fp_t Pmax, fp_t Psd)
{
  int i, j;
  int t, s;
  int x, y;
  graph_t g = graph_alloc(n*m);
  
  /* These nested loops connect up the basic lattice.
   */
  for (i = 0; i < n; i++) {
    for (j = 0; j < m; j++) {
      s = (i*m)+j;
      for (x=i-1;x<=i+1;x++) {
	for (y=j-1;y<=j+1;y++) {
	  if (x >= 0 && y >= 0 && x < n && y < m) {
	    t = (x*m)+y;
	    if (s != t) {
	      if (Psd >= 0.0) {
		graph_add_edge(g, s, t, (Psd * rnd_nrm()) + Pmax);
	      }
	      else {
		graph_add_edge(g, s, t, Pmax);
	      }
	    }
	  }
	}
      }
    }
  }

  return smallworld(g, p);
}

graph_t line(int n, int k, fp_t p, fp_t Pmax, fp_t Psd)
{
  int i, j, t;
  graph_t g = graph_alloc(n);

  /* Don't crash in the unlikely event that n < k
   */
  if (n < k) {
    k = n - 1;
    if (k < 0 || n < 0) {
      return NULL;
    }
  }

  /* These nested loops connect up the basic lattice.
   */
  for (i = 0; i < n; i++) {
    for (j = i-(k/2); j <= i+(k/2); j++) {
      if (j < 0 || j >= n) 
	continue;

      t = j;
      
      if (t != i) {
	if (Psd >= 0.0) {
	  graph_add_edge(g, i, t, (Psd * rnd_nrm()) + Pmax);
	}
	else {
	  graph_add_edge(g, i, t, Pmax);
	}
      }
    }
  }
  return smallworld(g, p);
}

graph_t twolayer(int n, int k, fp_t p, char *inh_p)
{
  /* "Two-layer" graph, after Bazhenov 2004. Basically each excitatory
   * neuron connects to its k excitatory neighbors, and each
   * inhibitory neuron connects to its k excitatory
   * neighbors. However, excitatory cells only connect to two
   * inhibitory neighbors. There are no inh-inh connections.
   */

  int i, j, t;
  graph_t g = graph_alloc(n);

  int ne;                    /* number of excitatory neurons */
  int me[n];                 /* map from relative to absolute index */
  int ni;                    /* number of inhibitory neurons */
  int mi[n];                 /* map from relative to absolute index */
  int r;                     /* ratio of excitatory:inhibitory */

  /* Don't crash in the unlikely event that n < k
   */
  if (n < k) {
    k = n - 1;
    if (k < 0 || n < 0) {
      return NULL;
    }
  }

  /* First we need to count up all of the excitatory and inhibitory
   * neurons and figure out their mappings.
   */

  ne = 0;
  ni = 0;
  for (i = 0; i < n; i++) {
    if (inh_p[i]) {
      mi[ni++] = i;
    }
    else {
      me[ne++] = i;
    }
  }

  if (ni == 0) {
    r = 0;
  }
  else {
    r = ne/ni;
  }

  fprintf(stderr, "Found %d excitatory and %d inhibitory neurons\n", ne, ni);

  /* These nested loops connect up the basic lattice.
   */
  for (i = 0; i < ne; i++) {
    for (j = i-(k/2); j <= i+(k/2); j++) {
      if (j < 0 || j >= ne) 
        continue;

      t = j;
      
      if (t != i) {
        graph_add_edge(g, me[i], me[t], 1.0);
      }

    }
    if (ni != 0) {
      /* Add connections to two adjacent inhibitory neurons.
       */
      t = (i/r)-1;
      if (t > 0)
        graph_add_edge(g, me[i], mi[t], 0.0);

      t = (i/r);
      if (t < ni) {
        graph_add_edge(g, me[i], mi[t], 0.0);
      }
    }
  }

  for (i = 0; i < ni; i++) {
    int c = i*r;
    for (j = c-(k/2); j <= c+(k/2); j++) {
      if (j < 0 || j >= ne)
        continue;

      graph_add_edge(g, mi[i], me[j], -1.0);
    }
  }
  return g;
}

void graph_print(FILE *fp, graph_t g, char *inh_p)
{
  int i;
  edge_t e_p;

  fprintf(fp, "digraph Net {\n");

  for (i = 0; i < g->n; i++) {
    for (e_p = g->e[i]; e_p != NULL; e_p = e_p->link) {
      
      fprintf(fp, " %c%d -> %c%d;\n", 
              inh_p[e_p->src] ? 'i' : 'e', e_p->src, 
              inh_p[i] ? 'i' : 'e', i);
    }
  }

  fprintf(fp, "}\n");
}
