/* iafs.c - Network code, synapses, etc.
   (The odd name is an out-of-date reference to when this model was built
    with integrate and fire neurons)
   Author: Robert D. Vincent
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "mersenne.h"
#include "stdjunk.h"
#include "rnd.h"
#include "graph.h"
#include "iafs.h"
#include "ga.h"
#include "wb.h"

#define KO_EQ 3.5
#define NAI_EQ 10.0
#define NAO_EQ 130.0
#define KI_EQ 100.0

#define FARADAY 96485.3

/* Calculate the equilibrium potential for an ion, using the Nernst
 * equation. The return value is in volts, so we typically have to
 * convert to mV.
 */
fp_t
nernst(fp_t Ci,			/* internal concentration */
       fp_t Co,			/* external concentration */
       fp_t T,			/* temperature (degrees C) */
       int Z)			/* charge on ion */
{
  /* 8.31451 is the universal gas constant */
  return (8.31451 * (T + 273.15)) / (Z * FARADAY) * log(Co / Ci);
}

#define _OFFSETOF(strname,fldname) (int)&(((struct strname *)0)->fldname)

#define _FDECL(name,type) { #name, _OFFSETOF(hh_net, name), type }

struct field {
  char name[12];
  short offset;
  short type;
};

struct field hh_fields[] = {
  _FDECL(n, T_INT),
  _FDECL(dT,T_FLOAT),
  _FDECL(Vth,T_FLOAT),
  _FDECL(nsd,T_FLOAT),
  _FDECL(nmu,T_FLOAT),
  _FDECL(Tr,T_FLOAT),
  _FDECL(EL,T_FLOAT),
  _FDECL(EK,T_FLOAT),
  _FDECL(ENa,T_FLOAT),
  _FDECL(cm,T_FLOAT),
  _FDECL(Pmax,T_FLOAT),
  _FDECL(Psd,T_FLOAT),
  _FDECL(gK,T_FLOAT),
  _FDECL(gL,T_FLOAT),
  _FDECL(gNa,T_FLOAT),
  _FDECL(TX,T_FLOAT),
  _FDECL(dX,T_FLOAT),
  _FDECL(dD,T_FLOAT),
  _FDECL(EAMPA,T_FLOAT),
  _FDECL(gAMPA,T_FLOAT),
  _FDECL(EGABAA,T_FLOAT),
  _FDECL(gGABAA,T_FLOAT),
  _FDECL(gGABAB,T_FLOAT),
  _FDECL(CoMg,T_FLOAT),
  _FDECL(degC,T_FLOAT),
  _FDECL(ninh,T_INT),
  _FDECL(gNaPx,T_FLOAT),
  _FDECL(gNax,T_FLOAT),
  _FDECL(gKdrx, T_FLOAT),
  _FDECL(gKAx, T_FLOAT),
  _FDECL(gKslowx, T_FLOAT),
  _FDECL(gLx, T_FLOAT),
  _FDECL(gNMDA, T_FLOAT),
  _FDECL(ENMDA, T_FLOAT),
  _FDECL(Bmax, T_FLOAT),
  _FDECL(Imax, T_FLOAT),
  _FDECL(Ko_th, T_FLOAT),
  _FDECL(Xinit, T_FLOAT),
  _FDECL(Prate, T_FLOAT),
  {"",-1,-1}
};

/* Calculate effective NMDA conductance using model of Jahr and Stevens (1990).
 */
fp_t GNMDA(fp_t v, fp_t CoMg)
{
  return  1.0 / (1.0 + CoMg/3.57 * exp(-v/16.13));
}

fp_t KNMDA(fp_t Ko)
{
  /* 1% at 3.5, 99% at 20 */
  return 100.0 / (1.0 + exp(-(Ko - 11.75) / 1.7974));
}

/*** Possible alternative implementations of GNMDA()
 **/
#if 0
/* Calculate effective NMDA conductance using model used by Bazhenov
 */
fp_t GNMDA(fp_t v, fp_t CoMg)
{
  return 1.0 / (1.0 + exp(-(v + 25.0)/12.5));
}
#endif
#if 0
/* Calculate effective NMDA conductance using model used by Golomb 2006
 */
fp_t GNMDA(fp_t v, fp_t CoMg)
{
  return 1.0 / (1.0 + exp(-(v + 50.0)/10.0));
}
#endif

/* Cell area in most of these models is 10^(-6) cm^2
 */
struct hh_net *
init_hh(const char *opts_str)
{
  struct hh_net *net_p = (struct hh_net *) malloc(sizeof(struct hh_net));
  int i;
  const char *key_str, *val_str, *tmp_str;
  int key_len;
  fp_t ftmp;
  int itmp;

  net_p->n = 1000;		/* Number of neurons */
  net_p->dT = 0.001;		/* Time step (seconds) */
  net_p->Vth = 0;		/* threshold potential, mV */
  net_p->Pmax = 4.0;		/* Maximum postsynaptic potential, nA */
  net_p->Psd = 0.0;		/* Standard deviation of psp, nA */
  net_p->nsd = 0.0;		/* noise standard deviation, nA */
  net_p->nmu = 0.0;		/* noise mean, nA */
  net_p->EAMPA = 0.0;		/* AMPA reversal potential, mV */
  net_p->gAMPA = 0.02;		/* Max. AMPA conductance, mS/cm^2 */
  net_p->ENMDA = 0.0;
  net_p->gNMDA = 0.01;
  net_p->gGABAA = 0.01;
  net_p->gGABAB = 0.01;
  net_p->EGABAA = -75;

  net_p->CoMg = 1.2;		/* extracellular Mg2+, mM */
  net_p->TX = 24;		/* K+ recovery time constant, seconds */
  net_p->dX = 0.0032;

  net_p->Tr = 0.7;		/* deactivation recovery time, seconds */
  net_p->dD = 0.07;		/* "resources" used per firing */

  net_p->cm = 1.0;		/* Membrane capacitance, uF/cm^2 */
  net_p->degC = 37;

  /* Parameters for Wang-Buzsaki FS cells 
   */
  net_p->gL = 0.1;		/* Leak conductance, mS/cm^2 */
  net_p->EL = -65;		/* Reversal potential of leak current, mV */
  net_p->gK = 9;		/* K+ conductance, mS/cm^2 */
  net_p->EK = -90.0;
  net_p->gNa = 35;		/* Na+ conductance, mS/cm^2 */
  net_p->ENa = 55.0;		/* Reversal potential of Na+ current, mV */

  /* Parameters for Golomb-style PY cells 
   */
  net_p->gNaPx = 0.06;		/* Persistent Na+ conductance, mS/cm^2 */
  net_p->gNax = 35.0;		/* Fast Na+ conductance, mS/cm^2 */
  net_p->gKdrx = 6.0;		/* Delayed rectifier K+ */
  net_p->gKAx = 0.0;		/* K+ A-type - not used! */
  net_p->gKslowx = 1.8;		/* Slow K+ current, mS/cm^2 */
  net_p->gLx = 0.05;		/* Leak conductance, mS/cm^2 */

  net_p->Bmax = 500;
  net_p->Imax = 10;
  net_p->Ko_th = 15;

  net_p->Xinit = 0.60;		/* Initial value of X[] */

  net_p->Prate = 1.0/8.0;	/* Poisson firing rate */

  for (key_str = opts_str;
       key_str != NULL && (tmp_str = strchr(key_str, '=')) != NULL;
       key_str = tmp_str) {
    key_len = tmp_str - key_str;
    val_str = ++tmp_str;
    tmp_str = strchr(val_str, ',');
    if (tmp_str != NULL) {
      tmp_str++;
    }
    for (i = 0; hh_fields[i].offset >= 0; i++) {
      int name_len = strlen(hh_fields[i].name);
      if (name_len == key_len &&
	  !strncmp(hh_fields[i].name, key_str, key_len)) {
	switch (hh_fields[i].type) {
	case T_INT:
	  itmp = atoi(val_str);
	  *((int *)((char *)net_p + hh_fields[i].offset)) = itmp;
	  fprintf(stderr, "%s=%d\n", hh_fields[i].name, itmp);
	  break;
	case T_FLOAT:
	  ftmp = atof(val_str);
	  *((fp_t *)((char *)net_p + hh_fields[i].offset)) = ftmp;
	  fprintf(stderr, "%s=%f\n", hh_fields[i].name, ftmp);
	  break;
	}
	break;
      }
    }
    if (hh_fields[i].offset < 0) {
      fprintf(stderr, "Unknown option name '%.*s'\n", key_len, key_str);
    }
  }

  net_p->Im = (fp_t *) malloc(sizeof(fp_t) * net_p->n);
  net_p->chvars = (fp_t *) malloc(sizeof(fp_t) * N_CHVARS * net_p->n);

  net_p->active = (char *) malloc(sizeof(char) * net_p->n);
  net_p->Tlast = (fp_t *) malloc(sizeof(fp_t) * net_p->n);
  net_p->F = (char *) malloc(sizeof(char) * net_p->n);
  net_p->D = (fp_t *) malloc(sizeof(fp_t) * net_p->n);
  net_p->Isyn = (fp_t *) malloc(sizeof(fp_t) * net_p->n);
  net_p->inh = (char *) malloc(sizeof(char) * net_p->n);
  net_p->O = (fp_t *) malloc(sizeof(fp_t) * net_p->n);
  net_p->OB = (fp_t *) malloc(sizeof(fp_t) * net_p->n);
  net_p->X = (fp_t *) malloc(sizeof(fp_t) * net_p->n);

  for (i = 0; i < net_p->n*N_CHVARS; i++) {
    net_p->chvars[i] = 0.0;
  }

  for (i = 0; i < net_p->n; i++) {
    fp_t *chvars = &net_p->chvars[i * N_CHVARS];
    net_p->Im[i] = 0.0;
    net_p->inh[i] = (char)((i % net_p->ninh) == 0);
    if (net_p->inh[i]) {
      init_wb(chvars);

    }
    else {
      init_ga(chvars);
    }

    chvars[I_GA] = 0.0;			/* GABA_A/AMPA r */
    chvars[I_GB1] = 0.0;		/* GABA_B/NMDA rb */
    chvars[I_GB2] = 0.0;		/* GABA_B sb */
    chvars[I_B] = 499.9494;
    chvars[I_CoK] = KO_EQ;
    chvars[I_CiK] = KI_EQ;
    chvars[I_CiNa] = NAI_EQ;
    chvars[I_CoNa] = NAO_EQ;

    net_p->active[i] = 1;
    net_p->Tlast[i] = -1000.0;
    net_p->F[i] = 0;
    net_p->D[i] = 1.0;
    net_p->O[i] = 0.0;
    net_p->X[i] = net_p->Xinit;
  }

  init_exc_act(net_p);
  init_inh_act(net_p);

  net_p->phi = pow(5, (net_p->degC - 27) / 10);

  return net_p;
}

void
free_hh(struct hh_net *net_p)
{
  if (net_p != NULL) {
    free(net_p->Im);
    free(net_p->chvars);
    free(net_p->Tlast);
    free(net_p->F);
    free(net_p->D);
    free(net_p->active);
    free(net_p->Isyn);
    free(net_p->O);
    free(net_p->OB);
    free(net_p->X);
    free(net_p);
  }
}

int _x = 0;

/* Basic sigmoid function. Domain is -inf to +inf, range is 0 to 1.
 * 't' sets the half-activation point (sigmoid(x,t,s)=0.5 iff x=t) and
 * 's' sets the "slope" of the curve.
 */
fp_t 
sigmoid(fp_t x, fp_t t, fp_t s)
{
  return (fp_t)(1.0 / (1.0 + exp(-(x - t) / s)));
}

void
check_fired(struct hh_net *net_p, int i, fp_t Tnow, fp_t v_prev)
{
  /* See if the membrane voltage has crossed the firing threshold, or
   * if our Poisson train should fire.
   */
  if ((v_prev < net_p->Vth && net_p->chvars[i*N_CHVARS+I_Vm] > net_p->Vth) ||
      (rnd_uni() < (net_p->dT*net_p->Prate))) {

    net_p->D[i] *= (1.0 - net_p->dD);
    net_p->X[i] *= (1.0 - net_p->dX);
    net_p->Tlast[i] = Tnow;
    net_p->F[i] = 1;
  }
  else {
    net_p->D[i] += ((1.0 - net_p->D[i]) * net_p->dT) / net_p->Tr;
    net_p->X[i] += ((1.0 - net_p->X[i]) * net_p->dT) / net_p->TX;
    net_p->F[i] = 0;
  }
}

#define TPULSE 0.001
#define TMAX 1.0e-3

/* Constants for AMPA synapses, from Destexhe, Mainen and Sejnowski
 * 1998.
 */
#define alpha_a 1.1e6		/* M^(-1) sec */
#define beta_a 190.0		/* sec^(-1) */
#define a_inf ((alpha_a * TMAX) / ((alpha_a * TMAX) + beta_a))
#define tau_a (1.0 / ((alpha_a * TMAX) + beta_a))

/* Constants for NMDA synapses.
 */
#define alpha_n 7.2e4 /* M^(-1) sec */
#define beta_n 6.6    /* sec^(-1) */
#define n_inf ((alpha_n * TMAX) / ((alpha_n * TMAX) + beta_n))
#define tau_n (1.0 / ((alpha_n * TMAX) + beta_n))


void
init_exc_act(struct hh_net *net_p)
{
  net_p->a_growth = exp(-net_p->dT / tau_a);
  net_p->a_decay = exp(-net_p->dT * beta_a);
  net_p->n_growth = exp(-net_p->dT / tau_n);
  net_p->n_decay = exp(-net_p->dT * beta_n);
}

/* Constants for GABA A synapses 
 */
#define alpha_r 5.0e6		/* M^(-1) sec^(-1) */
#define beta_r 180.0		/* sec^(-1) */
#define r_inf ((alpha_r * TMAX) / ((alpha_r * TMAX) + beta_r))
#define tau_r (1.0 / ((alpha_r * TMAX) + beta_r))

/* Constants for GABA B synapses */
#define Kd 100.0			// uM^4
#define K1 9.0e4
#define K2 1.2
#define K3 180.0
#define K4 34.0
#define rb_inf ((K1 * TMAX) / ((K1 * TMAX) + K2))
#define tau_rb (1.0 / ((K1 * TMAX) + K2))

void
init_inh_act(struct hh_net *net_p)
{
  net_p->r_growth = exp(-net_p->dT / tau_r);
  net_p->r_decay = exp(-net_p->dT * beta_r);
  net_p->rb_growth = exp(-net_p->dT / tau_rb);
  net_p->rb_decay = exp(-net_p->dT * K2);
}

/*
 * Update the synaptic activation for excitatory neurons. This calculates
 * the appropriate activation for AMPA and NMDA.
 */
void
update_exc_act(struct hh_net *net_p, /* The network state and parameters */
	       int i,		     /*  The neuron index */
	       fp_t Tnow)	     /* The current time in seconds */
{
  fp_t *chvars = &net_p->chvars[i * N_CHVARS];
  fp_t a = chvars[I_AMPA];	/* Get AMPA activation */
  fp_t n = chvars[I_NMDA];	/* Get NMDA activation */

  /* Update the synaptic activation. This update rule essentially
   * integrates the value during the initial "on" pulse, then allows
   * the activation to decay exponentially. This has the virtue of
   * being relatively simple, and also automatically giving us
   * synaptic facilitation if we pick the constants right.
   *
   * This whole approach is straight out of Destexhe, Mainen and
   * Sejnowski "Kinetic Models of Synaptic Transmission" (1998) in
   * Methods of Neuronal Modeling, Koch and Segev eds. (p. 23)
   */
  if ((Tnow - net_p->Tlast[i]) < TPULSE) {
    a = a_inf + (a - a_inf) * net_p->a_growth;
    n = n_inf + (n - n_inf) * net_p->n_growth;
  }
  else {
    a *= net_p->a_decay;
    n *= net_p->n_decay;
  }

  chvars[I_AMPA] = a;
  net_p->O[i] = a;

  chvars[I_NMDA] = n;
  net_p->OB[i] = n;
}

/*
 * Update the synaptic activation for inhibitory neurons. This calculates
 * the appropriate activation for GABA_A and GABA_B
 */
void
update_inh_act(struct hh_net *net_p, /* The network state and parameters */
	       int i, 		/*  The neuron index */
	       fp_t Tnow)	/* The current time in seconds */
{
  fp_t *chvars = &net_p->chvars[i * N_CHVARS];
  fp_t r = chvars[I_GA];	/* Get GABA A synaptic activation */
  fp_t rb = chvars[I_GB1];	/* Get GABA B synaptic activation */
  fp_t sb = chvars[I_GB2];	/* Get GABA B synaptic activation */

  /* Calculate the synaptic activations, following the scheme
   * suggested by Destexhe et al. 1998.
   */

  sb = sb + net_p->dT * (K3 * rb - K4 * sb); 

  /* Update the synaptic activation. This update rule essentially
   * integrates the value during the initial "on" pulse, then allows
   * the activation to decay exponentially. This has the virtue of
   * being relatively simple, and also automatically giving us 
   * synaptic facilitation if we pick the constants right.
   * 
   */
  if ((Tnow - net_p->Tlast[i]) < TPULSE) {
    r = r_inf + (r - r_inf) * net_p->r_growth;
    rb = rb_inf + (rb - rb_inf) * net_p->rb_growth;
  }
  else {
    r *= net_p->r_decay;
    rb *= net_p->rb_decay;
  }

  chvars[I_GA] = r;
  chvars[I_GB1] = rb;
  chvars[I_GB2] = sb;

  /* For GABA A or AMPA */
  net_p->O[i] = r;

  /* For GABA B */
  net_p->OB[i] = pow(sb, 4) / (pow(sb, 4) + Kd);
}

/*
 * update extracellular and buffered potassium concentrations.
 */
void update_kna(struct hh_net *net_p, 
                fp_t *y, 
                int i, 
                fp_t ik, 
                fp_t ina, 
                fp_t *CoK_prev,
                fp_t *dy)
{
  const double k1 = 0.015;
  const double k1N = 1.1;
  const double Dx2 = 2.0e-6/0.0001;	/* diffusion from adjacent cells */
  const double d = 0.15;
  const double depth = 1.0;
  const double factor = 10.0 / (FARADAY * depth); 

  double Kr, Kl;
  int ir, il;
  double Apump;
  double iKpump;
  double iNapump;
  double k2;
  double driveK;
  double driveNa;
  double G;
  double fKo;
  double fKi;
  double fB;
  double fNai;
  double fNao;

  double Ko;
  double B;
  double Nai;

  double Imax;

  Ko = y[I_CoK];
  B = y[I_B];
  Nai = y[I_CiNa];

  /* Get the adjacent cells' K+ concentrations.
   */
  ir = i + 1;
  if (ir >= net_p->n) {
    ir = net_p->n - 1;
  }
  Kr = CoK_prev[ir];

  il = i - 1;
  if (il < 0) {
    il = 0;
  }
  Kl = CoK_prev[il];

  if (net_p->inh[i]) {
    Imax = 0.00016;
  }
  else {
    Imax = 0.00064487;
  }

  /* The pump acts as a restoring force, pushing the concentration back
   * towards the equilibrium.
   */

  Apump = 1.0 / (pow(1.0 + (KO_EQ / Ko), 2) * pow(1.0 + (NAI_EQ / Nai), 3));
  
  iKpump =  -2*Imax*Apump;
  iNapump = 3*Imax*Apump;

  ik += iKpump;
  ina += iNapump;
  
  k2 = k1 / (1.0 + exp((Ko - net_p->Ko_th) / (-1.09) /*(-1.15)*/));
  
  driveK = -factor * ik;
  driveNa = -factor * ina;

  fNai = driveNa;
  fNao = -driveNa/d;
  //fNai = 0;
  //fNao = 0;

#define TAU_CONC 2000.0

  fNai += (NAI_EQ - y[I_CiNa]) / TAU_CONC;
  fNao += (NAO_EQ - y[I_CoNa]) / TAU_CONC;

  G = (k1 * (net_p->Bmax - B) / k1N) - (k2 * Ko * B);

  fKi = driveK;
  fKo = -driveK / d + G + Dx2 * (Kl + Kr - 2.0 * Ko);

  //fKi = 0;
  //fKo = 0;

  fKi += (KI_EQ - y[I_CiK]) / TAU_CONC;
  fKo += (KO_EQ - y[I_CoK]) / TAU_CONC;

  fB = k1 * (net_p->Bmax - B) - (k2 * Ko * B);

  dy[I_CoNa] = fNao;
  dy[I_CiNa] = fNai;
  dy[I_CoK] = fKo;
  dy[I_CiK] = fKi;
  dy[I_B] = fB;
}
  
void 
ode_hh(int n, 
       struct hh_net *net_p, 
       fp_t x, 
       fp_t *y, 
       fp_t *dy, 
       fp_t *CoK_prev, 
       fp_t *Ie,
       graph_t S)
{
  int i;
  fp_t ik;
  fp_t ina;

  for (i = 0; i < net_p->n; i++) {
    struct edge *e;
    fp_t g_sum = 0.0;
    fp_t g_syn;
    fp_t E_syn;
    fp_t i_syn = 0.0;
    fp_t i_inf = Ie[i];		/* Get input current */
    fp_t *uy = &y[i * N_CHVARS];
    fp_t *udy = &dy[i * N_CHVARS];

    fp_t v = uy[I_Vm];	/* Get membrane voltage */

    /* Calculate the equilibrium potential for K+, in mV 
     */
    net_p->EK = nernst(uy[I_CiK], uy[I_CoK], net_p->degC, 1) * 1000.0;
    net_p->ENa = nernst(uy[I_CiNa], uy[I_CoNa], net_p->degC, 1) * 1000.0;

#define P_NA 0.05
#define P_CL 0.45

    net_p->EL = nernst(uy[I_CiK] + (P_NA * uy[I_CiNa]) + (P_CL * 130.0),
		       uy[I_CoK] + (P_NA * uy[I_CoNa]) + (P_CL * 8.0),
		       net_p->degC, 1) * 1000.0;

    for (e = S->e[i]; e != NULL; e = e->link) {
      /* Figure out all the conductances of all synapses.
       */
      if (!net_p->inh[e->src] && !net_p->inh[i]) {
	/* E-E connection */
	g_syn = net_p->gAMPA * net_p->O[e->src];
	E_syn = net_p->EAMPA;
	g_syn *= net_p->D[e->src]; /* Synaptic depression */
	g_syn *= net_p->X[e->src];

	g_sum += g_syn;
	i_inf += g_syn * E_syn;
	i_syn += g_syn * (v - E_syn);

	g_syn = net_p->gNMDA * KNMDA(uy[I_CoK]) * GNMDA(v, net_p->CoMg) * net_p->OB[e->src];
	E_syn = net_p->ENMDA;
	g_syn *= net_p->D[e->src];
	g_syn *= net_p->X[e->src];
      }
      else if (!net_p->inh[e->src] && net_p->inh[i]) {
	/* E-I connection */
	g_syn = net_p->gAMPA * net_p->O[e->src];
	E_syn = net_p->EAMPA;
	//g_syn *= net_p->D[e->src]; /* Synaptic depression */
      }
      else {
	/* I-I or I-E connection */
	g_syn = net_p->gGABAA * net_p->O[e->src];
	E_syn = net_p->EGABAA;

	/* No synaptic depression for GABA */
	g_sum += g_syn;
	i_inf += g_syn * E_syn;
	i_syn += g_syn * (v - E_syn);

	g_syn = net_p->gGABAB * net_p->OB[e->src];
	E_syn = net_p->EK;
      }
      g_sum += g_syn;
      i_inf += g_syn * E_syn;
      i_syn += g_syn * (v - E_syn);
    }

    net_p->Isyn[i] = i_syn;

    if (net_p->inh[i]) {
      update_wb(net_p, uy, i, i_inf, g_sum, &ik, &ina, udy);
    }
    else {
      update_ga(net_p, uy, i, i_inf, g_sum, &ik, &ina, udy);
    }

    update_kna(net_p, uy, i, ik, ina, CoK_prev, udy);
  }
}


/*
 * Procedure for simple Runge-Kutta 4th order finite-difference integration
 */

typedef void (*ode_func_t)(int n, struct hh_net *net_p, fp_t x, fp_t *y,
                           fp_t *dy, fp_t *K, fp_t *Ie, graph_t S);

void 
rk4(int n, 
    ode_func_t fun,
    struct hh_net *net_p,
    fp_t dt, 
    fp_t x, 
    fp_t *y, 
    fp_t *dy, 
    fp_t *s, 
    fp_t *yk,
    fp_t *K,
    fp_t *Ie,
    graph_t S)
{
  int i;
  fp_t xk;
  
  for (i = 0; i < n; i++) {
    dy[i] = 0.0;
  }

  fun(n, net_p, x, y, dy, K, Ie, S);

  for (i = 0; i < n; i++) { 
    s[i] = dy[i];
    yk[i] = y[i] + (dt / 2.0) * dy[i];
  }

  xk = x + dt / 2.0;

  fun(n, net_p, xk, yk, dy, K, Ie, S);

  for (i = 0; i < n; i++) { 
    s[i] += 2.0 * dy[i]; 
    yk[i] = y[i] + (dt / 2.) * dy[i]; 
  }

  fun(n, net_p, xk, yk, dy, K, Ie, S);

  for (i = 0; i < n; i++) { 
    s[i] += 2.0 * dy[i]; 
    yk[i] = y[i] + dt * dy[i];
  }

  xk = x + dt;
  fun(n, net_p, xk, yk, dy, K, Ie, S);

  for (i = 0; i < n; i++) { 
    y[i] += (dt / 6.0) * (s[i] + dy[i]);
  }
}

void
step_hh(struct hh_net *net_p, graph_t S, fp_t Tnow, fp_t *Ie)
{
  int i;
  fp_t step = net_p->dT * 1000.0;	/* convert from sec to msec */
  fp_t vm_prev[net_p->n];
  fp_t CoK_prev[net_p->n];
  fp_t dy[net_p->n * N_CHVARS];
  fp_t s[net_p->n * N_CHVARS];
  fp_t yk[net_p->n * N_CHVARS];

  if (net_p->nsd != 0.0) {
    for (i = 0; i < net_p->n; i++) {
      Ie[i] += (rnd_nrm() * net_p->nsd) + net_p->nmu;
    }
  }
  else if (net_p->nmu != 0.0) {
    for (i = 0; i < net_p->n; i++) {
      Ie[i] += net_p->nmu;
    }
  }

  for (i = 0; i < net_p->n; i++) {
    fp_t *chvars = &net_p->chvars[i * N_CHVARS];
    vm_prev[i] = chvars[I_Vm];
    CoK_prev[i] = chvars[I_CoK];
  }

  rk4(net_p->n * N_CHVARS, 
      ode_hh,
      net_p,
      step,
      0,                        /* x */
      net_p->chvars,
      dy,
      s,
      yk,
      CoK_prev,
      Ie,
      S);
      
  /* Perform a second pass in which we update all of the synaptic
   * activations and firing-related variables.
   */
  for (i = 0; i < net_p->n; i++) {
    check_fired(net_p, i, Tnow, vm_prev[i]);
    if (net_p->inh[i]) {
      update_inh_act(net_p, i, Tnow); /* Update synaptic activations */
    }
    else {
      update_exc_act(net_p, i, Tnow);
    }
  }
}
