/* wb.c - Wang/Buzsaki cell 
 * Author: Robert D. Vincent
 */
#include <math.h>
#include <stdio.h>

#include "stdjunk.h"
#include "graph.h"
#include "iafs.h"
#include "wb.h"

/* Resting values for Wang-Buzsaki model */
#define V0 -63.808

void
init_wb(fp_t *y)
{
  fp_t alpha_n = 0.01 * (V0 + 34) / (1 - exp(-(V0 + 34) / 10));
  fp_t beta_n = 0.125 * exp(-(V0 + 44) / 80);
  fp_t n_inf = alpha_n / (alpha_n + beta_n);
  fp_t alpha_h = 0.07 * exp(-(V0 + 58.0) / 20.0);
  fp_t beta_h = 1.0 / (1 + exp(-(V0 + 28.0) / 10.0));
  fp_t h_inf = alpha_h / (alpha_h + beta_h);
  y[I_Vm] = V0;
  y[I_USER+0] = n_inf;
  y[I_USER+1] = h_inf;
}

void
update_wb(struct hh_net *net_p, 
          fp_t *y,
          int i, 
          fp_t i_inf, 
          fp_t g_sum, 
          fp_t *ik, 
          fp_t *ina, 
          fp_t *dy)
{
  fp_t v = y[I_Vm];
  fp_t n = y[I_USER+0];
  fp_t h = y[I_USER+1];
  fp_t i_m = 0;
  fp_t tau_v, v_inf;
  fp_t tau_n, n_inf;
  fp_t tau_h, h_inf;
  fp_t m_inf;
  fp_t alpha_m, beta_m;
  fp_t alpha_n, beta_n;
  fp_t alpha_h, beta_h;

  fp_t g_tmp;
  fp_t i_tmp;

  const fp_t E_L = -65; //net_p->EL;

  *ik = 0;
  *ina = 0;

  /* Add the leak conductance to the total conductance. 
   */
  g_tmp = net_p->gL;
  g_sum += g_tmp;
  i_inf += g_tmp * E_L;
  i_tmp = g_tmp * (v - E_L);
  i_m += i_tmp;

  g_tmp = net_p->gK * pow(n, 4);
  g_sum += g_tmp;
  i_inf += g_tmp * net_p->EK;
  i_tmp = g_tmp * (v - net_p->EK);
  i_m += i_tmp;
  *ik += i_tmp;

  alpha_m = 0.1 * (v + 35.0) / (1 - exp(-(v + 35.0) / 10.0));
  beta_m = 4.0 * exp(-(v + 60.0) / 18.0);
   
  m_inf = alpha_m / (alpha_m + beta_m);

  g_tmp = net_p->gNa * pow(m_inf, 3) * h;
  g_sum += g_tmp;
  i_inf += g_tmp * net_p->ENa;
  i_tmp = g_tmp * (v - net_p->ENa);
  i_m += i_tmp;
  *ina += i_tmp;

  /* Calculate the time constant and steady-state value for Vm */
  tau_v = net_p->cm / g_sum;
  v_inf = i_inf / g_sum;

  /* Calculate alpha and beta for the delayed-rectifier K+
   * conductance. This is what we consider a "persistent"
   * conductance, in that it is gated by a single process
   * that is voltage dependent.
   */
  alpha_n = net_p->phi * 0.01 * (v + 34) / (1 - exp(-(v + 34) / 10));
  beta_n = net_p->phi * 0.125 * exp(-(v + 44) / 80);
   
  tau_n = 1.0 / (alpha_n + beta_n);
  n_inf = alpha_n / (alpha_n + beta_n);
   
  /* Calculate alpha and beta for both components of the fast Na+
   * conductance. This is a "transient" conductance, in that it is
   * controlled by two processes with opposite voltage dependencies.
   */
  alpha_h = net_p->phi * 0.07 * exp(-(v + 58.0) / 20.0);
  beta_h = net_p->phi * 1.0 / (1 + exp(-(v + 28.0) / 10.0));
  
  tau_h = 1.0 / (alpha_h + beta_h);
  h_inf = alpha_h / (alpha_h + beta_h);

  /* Integrate
   */

  dy[I_Vm] = (v_inf - v) / tau_v;
  dy[I_USER+0] = (n_inf - n) / tau_n;
  dy[I_USER+1] = (h_inf - h) / tau_h;

  net_p->Im[i] = i_m;
}

